from statsmodels.base.model import GenericLikelihoodModel
import numpy as np
import scipy.stats as scs

class LogProbitSig(GenericLikelihoodModel):
    def __init__(self, endog, exog, weights):
        self.weights = weights
        super(LogProbitSig, self).__init__(endog, exog)

    def loge(self, params):
        pred = np.dot(self.exog, np.exp(params[:-1]))
        return scs.norm.logcdf(np.log(pred), scale=np.exp(params[-1]))

    def log1e(self, params):
        pred = np.dot(self.exog, np.exp(params[:-1]))
        return scs.norm.logsf(np.log(pred), scale=np.exp(params[-1]))
    
    def loglike(self, params):
        endog = self.endog
        return (self.weights*endog*self.loge(params) + (1-endog)*self.weights*self.log1e(params)).sum()
    
    def fit(self, start_params=None, maxiter=10000, maxfun=5000, **kwds):
        # we have one additional parameter and we need to add it for summary
        self.exog_names.append('sig')
        if start_params is None:
            # Reasonable starting values
            # start_params = np.zeros((self.exog.shape[1] + 1))
            start_params = -0.5 + np.random.rand(self.exog.shape[1] + 1)
        return super(LogProbitSig, self).fit(start_params=start_params,
                                     maxiter=maxiter, maxfun=maxfun,
                                     **kwds)
class Gumbel(GenericLikelihoodModel):
    def __init__(self, endog, exog, weights):
        self.weights = weights
        super(Gumbel, self).__init__(endog, exog)

    def loge(self, params):
        pred = np.dot(self.exog, np.exp(params[:-1]))
        return scs.weibull_min.logcdf(pred, np.exp(params[-1]))

    def log1e(self, params):
        pred = np.dot(self.exog, np.exp(params[:-1]))
        return scs.weibull_min.logsf(pred, np.exp(params[-1]))
    
    def loglike(self, params):
        endog = self.endog
        return (self.weights*endog*self.loge(params) + (1-endog)*self.weights*self.log1e(params)).sum()
    
    def fit(self, start_params=None, maxiter=10000, maxfun=5000, **kwds):
        # we have one additional parameter and we need to add it for summary
        self.exog_names.append('b')
        if start_params is None:
            # Reasonable starting values
            # start_params = np.zeros((self.exog.shape[1] + 1))
            start_params = -0.5 + np.random.rand(self.exog.shape[1] + 1)
        return super(Gumbel, self).fit(start_params=start_params,
                                     maxiter=maxiter, maxfun=maxfun,
                                     **kwds)


import numpy as np
import scipy.stats as scs
import statsmodels.api as sm
from .likelihood_models import LogProbitSig, Gumbel

class Propensity:
    '''
    Instance of propensity model to be used in a PU classification model
    '''
    def __init__(self):
        self.params = None
    
    def __str__(self):
        return str(self.params)
    
    def __repr__(self):
        return self.__str__()
    
    def initialization(self, Xe):
        self.params = np.random.rand(Xe.shape[1] + 1)
    
    def e(self, Xe):
        return 0.5*np.ones(Xe.shape[0])
    
    def loge(self, Xe):
        return np.zeros(Xe.shape[0])
    
    def predict_proba(self, Xe):
        return self.e(Xe)

    def predict_logproba(self, Xe):
        return self.loge(Xe)
    
    def predict(self, Xe, threshold=0.5):
        return (self.predict_proba(Xe)>=threshold).astype(int)

class LogProbitPropensity(Propensity):
    def __init__(self):
        return super(LogProbitPropensity, self).__init__()

    def __str__(self):
        if self.params is not None:
            s = 'Phi = {}'.format(self.params[:-1])
            s += '\n'
            s += 'sigma = {}'.format(self.params[-1])
            return s
        else:
            return ''

    def __repr__(self):
        return self.__str__()

    def initialization(self, Xe, gamma, Y, w=1.):
        self.params = np.random.rand(Xe.shape[1] + 1)
    
    def e(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return scs.norm.cdf(np.log(pred), scale=self.params[-1])
        

    def loge(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return np.max(np.concatenate([scs.norm.logcdf(np.log(pred), scale=self.params[-1])[:,np.newaxis], -1e99*np.ones((Xe.shape[0],1))], axis=1), axis=1)

    def log1e(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return scs.norm.logsf(np.log(pred), scale=self.params[-1])
    
    def fit(self, Xe, gamma, Y, w=1., warm_start=True, balance=False):
        if self.params is None:
            self.initialization(Xe, gamma, Y, w)
        if warm_start == False:
            start_params = None
        else:
            start_params = np.log(self.params.copy())
        if balance:
            q1, q2 = sum(Y)/sum(gamma), 1-sum(Y)/sum(gamma)
            c1, c2 = 1/q1/(1/q1 + 1/q2), 1/q2/(1/q1 + 1/q2)
            weights = c1*Y + c2*(1-Y)
        else:
            weights = 1
        self.model = LogProbitSig(Y, Xe, weights*gamma)
        self.res = self.model.fit(start_params=start_params, disp=0)
        self.params = np.exp(self.res.params.copy())

class LogisticPropensity(Propensity):
    '''
    Logistic Propensity : the feature vector Xe is assumed to contain an intercept as its first column
    '''
    def __init__(self):
        return super(LogisticPropensity, self).__init__()  

    def __str__(self):
        if self.params is not None:
            s = 'phi = {}'.format(self.params)
            return s
        else:
            return ''

    def __repr__(self):
        return self.__str__()

    def initialization(self, Xe, gamma, Y, w=1.):
        self.params = np.random.randn(Xe.shape[1]+1)
    
    def e(self, Xe):
        # add intercept
        Xe = np.concatenate([np.ones((Xe.shape[0],1)), Xe], axis=1)
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.dot(Xe, self.params)
        return scs.logistic.cdf(pred)
        

    def loge(self, Xe):
        # add intercept
        Xe = np.concatenate([np.ones((Xe.shape[0],1)), Xe], axis=1)
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.dot(Xe, self.params)
        return scs.logistic.logcdf(pred)

    def log1e(self, Xe):
        # add intercept
        Xe = np.concatenate([np.ones((Xe.shape[0],1)), Xe], axis=1)
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.dot(Xe, self.params)
        return scs.logistic.logsf(pred)
    
    def fit(self, Xe, gamma, Y, w=1., warm_start=True, balance=False):
        if self.params is None:
            self.initialization(Xe, gamma, Y, w)
        if warm_start == False:
            start_params = None
        else:
            start_params = self.params.copy()
        if balance:
            q1, q2 = sum(Y)/sum(gamma), 1-sum(Y)/sum(gamma)
            c1, c2 = 1/q1/(1/q1 + 1/q2), 1/q2/(1/q1 + 1/q2)
            weights = c1*Y + c2*(1-Y)
        else:
            weights = 1
        # add intercept
        Xe = np.concatenate([np.ones((Xe.shape[0],1)), Xe], axis=1)
        self.model = sm.GLM(Y, Xe, freq_weights=weights*gamma, family=sm.families.Binomial())
        self.res = self.model.fit(start_params=start_params, disp=0)
        self.params = self.res.params.copy()

class GumbelPropensity(Propensity):
    def __init__(self):
        return super(GumbelPropensity, self).__init__()

    def __str__(self):
        if self.params is not None:
            s = 'Phi = {}'.format(self.params[:-1])
            s += '\n'
            s += 'k = {}'.format(self.params[-1])
            return s
        else:
            return ''

    def __repr__(self):
        return self.__str__()

    def initialization(self, Xe, gamma, Y, w=1.):
        self.params = np.random.rand(Xe.shape[1] + 1)   

    def e(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return scs.weibull_min.cdf(pred, self.params[-1])

    def loge(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return np.max(np.concatenate([scs.weibull_min.logcdf(pred, self.params[-1])[:,np.newaxis], -1e99*np.ones((Xe.shape[0],1))], axis=1), axis=1)

    def log1e(self, Xe):
        if self.params is None:
            print('Please initialise parameters first')
            return
        pred = np.max(np.concatenate([np.dot(Xe, self.params[:-1])[:,np.newaxis], 1e-30*np.zeros((Xe.shape[0],1))], axis=1), axis=1)
        return scs.weibull_min.logsf(pred, self.params[-1])
    
    def fit(self, Xe, gamma, Y, w=1., warm_start=True, balance=False):
        if self.params is None:
            self.initialization(Xe, gamma, Y, w)
        if warm_start == False:
            start_params = None
        else:
            start_params = np.log(self.params.copy())
        if balance:
            q1, q2 = sum(Y)/sum(gamma), 1-sum(Y)/sum(gamma)
            c1, c2 = 1/q1/(1/q1 + 1/q2), 1/q2/(1/q1 + 1/q2)
            weights = c1*Y + c2*(1-Y)
        else:
            weights = 1
        self.model = Gumbel(Y, Xe, weights*gamma)
        self.res = self.model.fit(start_params=start_params, disp=0)
        # print(self.res.summary())
        self.params = np.exp(self.res.params.copy()) 


import statsmodels.api as sm
import numpy as np
import scipy.stats as scs

class Classifier:
    '''
    Instance of classifier to be used in a PU classification model
    '''
    def __init__(self):
        self.params = None

    def initialization(self, Xc):
        self.params = np.random.rand(Xc.shape[1] + 1)

    def eta(self, Xc):
        return 0.5*np.ones(Xc.shape[0])

    def logeta(self, Xc):
        return np.zeros(Xc.shape[0])

    def predict_proba(self, Xc):
        return self.eta(Xc)

    def predict_logproba(self, Xc):
        return self.logeta(Xc)

    def predict(self, Xc, threshold=0.5):
        return (self.predict_proba(Xc)>=threshold).astype(int)

class LinearLogisticRegression(Classifier):
    def __init__(self):
        super(LinearLogisticRegression, self).__init__()

    def __str__(self):
        if self.params is not None:
            s = 'theta_1 = {}'.format(self.params)
            return s
        else:
            return ''

    def __repr__(self):
        return self.__str__()

    def initialization(self, Xc, gamma, w=1.):
        self.params = np.random.randn(Xc.shape[1]+1)

    def fit(self, Xc, gamma, w=1., warm_start=True):
        # add intercept
        Xc = np.concatenate([np.ones((Xc.shape[0],1)), Xc], axis=1)
        if self.params is None:
            self.initialization(Xc, gamma, w)
        if warm_start == False:
            start_params = None
        else:
            start_params = self.params.copy()
        Xc_ = np.concatenate([Xc,Xc], axis=0)
        Z_ = np.concatenate([np.ones(len(gamma)),np.zeros(len(gamma))], axis=0)
        q1, q2 = sum(gamma)/len(gamma), 1-sum(gamma)/len(gamma)
        c1, c2 = 1/q1/(1/q1 + 1/q2), 1/q2/(1/q1 + 1/q2)
        w_ = np.concatenate([gamma, (1-gamma)], axis=0)
        self.model = sm.GLM(Z_, Xc_, freq_weights=w_, family=sm.families.Binomial())
        # self.model = LogisticRegression(penalty='none', max_iter=1e5)
        # else:
        #     self.model.exog = np.concatenate([Xc,Xc], axis=0)
        #     self.model.endog = np.concatenate([np.ones(len(gamma)),np.zeros(len(gamma))], axis=0)
        #     self.model.freq_weights = np.concatenate([w*gamma, w*(1-gamma)], axis=0)
        #     # params_init = self.params.copy()
        #     params_init = None
        self.res = self.model.fit(start_params=start_params, disp=0)
        # # else:
        # #     self.res = self.model.fit_regularized(alpha=self.C, start_params=start_params, maxiter=int(1e3))
        self.params = self.res.params.copy()
        # self.model.fit(Xc_, Z_, sample_weight=w_)
        # self.params = self.model.coef_[0]

    def eta(self, Xc):
        # add intercept
        Xc = np.concatenate([np.ones((Xc.shape[0],1)), Xc], axis=1)
        if self.params is None:
            print('Please initialize parameters first')
            return
        return scs.logistic.cdf(np.dot(Xc, self.params))

    def logeta(self, Xc):
        # add intercept
        Xc = np.concatenate([np.ones((Xc.shape[0],1)), Xc], axis=1)
        if self.params is None:
            print('Please initialize parameters first')
            return
        return scs.logistic.logcdf(np.dot(Xc, self.params))


class LinearDiscriminantClassifier(Classifier):
    def __init__(self):
        super(LinearDiscriminantClassifier).__init__()
    
    def initialization(self, Xc, gamma, w=1.):
        self.params = {'pi':np.random.rand(), 'mu_0':np.random.randn(Xc.shape[1]), 'mu_1':np.random.randn(Xc.shape[1]), 'Sigma':1/len(Xc)*np.dot((Xc-Xc.mean(axis=0)).T, Xc-Xc.mean())}
        # self.params = {'pi':np.mean(gamma), 'mu_0':np.mean(Xc[gamma==0,:], axis=0), 'mu_1':np.mean(Xc[gamma==1,:], axis=0), 'Sigma':1/len(Xc)*np.dot((Xc-Xc.mean(axis=0)).T, Xc-Xc.mean())}

        # self.Sigma = 1/len(Xc)*np.dot((Xc-Xc.mean(axis=0)).T, Xc-Xc.mean())
        # self.Sigma = np.eye(Xc.shape[1])
        # self.params = {'pi':gamma.mean(), 'mu_0':Xc[gamma==0,:].mean(axis=0), 'mu_1':Xc[gamma==1,:].mean(axis=0)}
        # X0, X1 = np.dot((Xc-self.params['mu_0']).T,(1-gamma[:,np.newaxis])*(Xc-self.params['mu_0'])), np.dot((Xc-self.params['mu_1']).T,gamma[:,np.newaxis]*(Xc-self.params['mu_1']))
        # self.params['Sigma'] = 1/len(Xc)*(X0+X1)



    def __str__(self):
        if self.params is not None:
            s = r'$\mu_0 = {}$'.format(self.params['mu_0'])
            s += r'$\mu_1 = {}$'.format(self.params['mu_1'])
            s += r'$\pi = {}$'.format(self.params['pi'])
            s += r'$\Sigma = {}$'.format(self.params['Sigma'])
            return s
        else:
            return ''

    def __repr__(self):
        return self.__str__()
    
    def fit(self, Xc, gamma, w=1., warm_start=True):
        self.params['pi'] = np.mean(gamma)
        self.params['mu_1'] = np.mean(gamma[:,np.newaxis]*w*Xc, axis=0)/np.mean(gamma)
        self.params['mu_0'] = np.mean((1-gamma)[:,np.newaxis]*w*Xc, axis=0)/np.mean(1-gamma)
        X0, X1 = np.dot((Xc-self.params['mu_0']).T,(1-gamma[:,np.newaxis])*(Xc-self.params['mu_0'])), np.dot((Xc-self.params['mu_1']).T,gamma[:,np.newaxis]*(Xc-self.params['mu_1']))
        self.params['Sigma'] = 1/len(Xc)*(X0+X1)
    
    def eta(self, Xc):
        f1 = scs.multivariate_normal.pdf(Xc, mean=self.params['mu_1'], cov=self.params['Sigma'])
        f0 = scs.multivariate_normal.pdf(Xc, mean=self.params['mu_0'], cov=self.params['Sigma'])
        return self.params['pi']*f1/(self.params['pi']*f1 + (1-self.params['pi'])*f0)
    
    def logeta(self, Xc):
        p0 = (np.dot((Xc-self.params['mu_0']), np.linalg.inv(self.params['Sigma']))*(Xc-self.params['mu_0'])).sum(axis=1)
        p1 = (np.dot((Xc-self.params['mu_1']), np.linalg.inv(self.params['Sigma']))*(Xc-self.params['mu_1'])).sum(axis=1)
        x = -np.log((1-self.params['pi'])/self.params['pi']) + 0.5 * p0 - 0.5 * p1
        return scs.logistic.logcdf(x)
    
    def pdf_pos(self, Xc):
        return self.params['pi']*scs.multivariate_normal.pdf(Xc, mean=self.params['mu_1'], cov=self.params['Sigma'])

    def pdf_neg(self, Xc):
        return (1-self.params['pi'])*scs.multivariate_normal.pdf(Xc, mean=self.params['mu_0'], cov=self.params['Sigma'])


"""Main module."""
import numpy as np
import pickle

class PUClassifier:
    '''
    PU learning classification model under unknown propensity.
    This model works by specifying a model on the classification and on the propensity and estimates parameters using EM algorithm (SAR-EM, Bekker et al.)
    '''
    def __init__(self, cmodel, emodel, pdf=False, da=False):
        '''
        Initialization of PU classifier. Two parameters are required:
        - cmodel: an instance of classification model
        - emodel: an instance of propensity model
        '''
        self.cmodel = cmodel
        self.emodel = emodel
        self.pdf = pdf
        self.da = da
        self.history = []
    
    def initialization(self, Xc, Xe, Y, gamma, w=1.):
        '''
        Initialization of parameters for both model before running EM algorithm
        '''
        self.cmodel.initialization(Xc, Y, w)
        self.emodel.initialization(Xe, gamma, Y, w)
        self.history = [self.loglikelihood(Xc, Xe, Y)]

    
    def __str__(self):
        s = 'PU classifier \n'
        s += self.cmodel.__str__()
        s += '\n'
        s += self.emodel.__str__()
        return s

    def __repr__(self):
        return self.__str__()
    
    def eta(self, Xc):
        return self.cmodel.eta(Xc)

    def logeta(self, Xc):
        return self.cmodel.logeta(Xc)

    def e(self, Xe):
        return self.emodel.e(Xe)

    def loge(self, Xe):
        return self.emodel.loge(Xe)
    
    def logp(self, Xe):
        # try:
        return self.emodel.logp(Xe)
        # except:
        #     print('Unavailable for this propensity model')

    def log1e(self, Xe):
        return self.emodel.log1e(Xe)
    
    def log1etae(self, Xc, Xe):
        leta, le = self.logeta(Xc), self.loge(Xe)
        approx = (leta + le <=-1e-10)
        return approx*np.nan_to_num(np.log1p(-np.exp(leta + le))) + (1-approx)*np.nan_to_num(np.log(-leta-le))
    
    def predict_cproba(self, Xc):
        return self.cmodel.eta(Xc)

    def predict_clogproba(self, Xc):
        return self.cmodel.logeta(Xc)
    
    def predict_c(self, Xc, threshold=0.5):
        return (self.predict_cproba(Xc)>=threshold).astype(int)

    def predict_eproba(self, Xe):
        return self.emodel.e(Xe)

    def predict_elogproba(self, Xe):
        return self.emodel.loge(Xe)
    
    def predict_e(self, Xe, threshold=0.5):
        return (self.predict_eproba(Xe)>=threshold).astype(int)
    
    def predict_proba(self, Xc, Xe):
        return self.predict_cproba(Xc)*self.predict_eproba(Xe)

    def predict_logproba(self, Xc, Xe):
        return self.predict_clogproba(Xc) + self.predict_elogproba(Xe)

    def predict(self, Xc, Xe, threshold=0.5):
        return (self.predict_proba(Xc, Xe)>=threshold).astype(int)
    
    def loglikelihood(self, Xc, Xe, Y, w=1.):
        if self.da:
            return self.loglikelihood_da(Xc, Xe, Y, w)
        else:
            return self.loglikelihood_lr(Xc, Xe, Y, w)
    
    def loglikelihood_lr(self, Xc, Xe, Y, w=1.):
        '''
        Compute the loglikelihood of the model given observations:
        - Xc: the feature vector used in the classification model
        - Xe: the feature vector used in the propensity model
        - Y: observed noisy PU labels. Y=1 (labeled) or Y=0 (unlabeled)
        - w: classification weights (default=1. else, should be an array of positive float with the same size as Y)
        '''
        if self.pdf == False:
            return np.mean(w*(Y*self.logeta(Xc) + Y*self.loge(Xe) + (1-Y)*self.log1etae(Xc, Xe)))
        else:
            return np.mean(w*(Y*self.logeta(Xc) + Y*self.logp(Xe) + (1-Y)*self.log1etae(Xc, Xe)))

    def loglikelihood_da(self, Xc, Xe, Y, w=1.):
        '''
        Compute the loglikelihood of the model given observations:
        - Xc: the feature vector used in the classification model
        - Xe: the feature vector used in the propensity model
        - Y: observed noisy PU labels. Y=1 (labeled) or Y=0 (unlabeled)
        - w: classification weights (default=1. else, should be an array of positive float with the same size as Y)
        '''
        return np.mean(Y*(np.log(self.cmodel.pdf_pos(Xc)) + self.emodel.loge(Xe)) + (1-Y)*np.log(self.cmodel.pdf_pos(Xc)*(1-self.emodel.e(Xe)) + self.cmodel.pdf_neg(Xc)))
    
    def expectation(self, Xc, Xe, Y):
        '''
        Compute the expectation step of EM algorithm, return the probabilities for every instance to be of positive class:
        - Xc: the feature vector used in the classification model
        - Xe: the feature vector used in the propensity model
        - Y: observed noisy PU labels. Y=1 (labeled) or Y=0 (unlabeled)
        '''
        l1 = self.logeta(Xc) + self.log1e(Xe)
        l2 = self.log1etae(Xc, Xe)
        p = np.min(np.concatenate([np.exp(l1-l2)[:,np.newaxis], np.ones((len(Y),1))], axis=1), axis=1)
        return Y + (1-Y)*p

    def maximisation(self, Xc, Xe, Y, gamma, w=1., warm_start=True, balance=False):
        '''
        Compute the maximisation step of EM algorithm, update the model parameters in both classification and propensity models
        - Xc: the feature vector used in the classification model
        - Xe: the feature vector used in the propensity model
        - Y: observed noisy PU labels. Y=1 (labeled) or Y=0 (unlabeled)
        - gamma: probabilities obtained at expectation step
        '''
        self.prev_params_c, self.prev_params_e = self.cmodel.params.copy(), self.emodel.params.copy()
        self.cmodel.fit(Xc, gamma, w, warm_start)
        self.emodel.fit(Xe, gamma, Y, w, warm_start, balance)
    
    def fit(self, Xc, Xe, Y, w=1., tol=1e-6, max_iter=1e4, warm_start=False, balance=False):
        if not warm_start:
            self.initialization(Xc, Xe, Y, Y, w)
        it = 0
        delta = 1
        last_ll = self.loglikelihood(Xc, Xe, Y, w)
        warm_start = False
        while abs(delta) > tol and it < max_iter:
            if delta<0:
                # print("Attention")
                self.cmodel.params = self.prev_params_c.copy()
                self.emodel.params = self.prev_params_e.copy()
            gamma = self.expectation(Xc, Xe, Y)
            self.maximisation(Xc, Xe, Y, gamma, w, warm_start, balance)
            delta = self.loglikelihood(Xc, Xe, Y, w) - last_ll
            last_ll = self.loglikelihood(Xc, Xe, Y, w)
            self.history.append(last_ll)
            # print(last_ll)
            it += 1
            warm_start = True
    
    def save(self, path):
        with open(path, 'wb') as f:
            pickle.dump(self, f)